import logging
from typing import List

import mkdocs
from mkdocs.utils import warning_filter
from mkdocs.structure import files as fl
from .file_mapper import FileMapper
from .replacer import EzLinksReplacer
from .scanners.md_link_scanner import MdLinkScanner
from .scanners.wiki_link_scanner import WikiLinkScanner
from .scanners.reference_link_scanner import ReferenceLinkScanner
from .types import EzLinksOptions

LOGGER = logging.getLogger(f"mkdocs.plugins.{__name__}")
LOGGER.addFilter(warning_filter)


class EzLinksPlugin(mkdocs.plugins.BasePlugin):
    config_scheme = (
        ("wikilinks", mkdocs.config.config_options.Type(bool, default=True)),
        ("warn_ambiguities", mkdocs.config.config_options.Type(bool, default=False)),
        ("reference_links", mkdocs.config.config_options.Type(bool, default=False)),
    )

    def init(self, config):
        self.replacer = EzLinksReplacer(
            root=config["docs_dir"],
            file_map=self.file_mapper,
            use_directory_urls=config["use_directory_urls"],
            options=EzLinksOptions(**self.config),
            logger=LOGGER,
        )

        self.replacer.add_scanner(MdLinkScanner())
        if self.config["wikilinks"]:
            self.replacer.add_scanner(WikiLinkScanner())

        if self.config["reference_links"]:
            self.replacer.add_scanner(ReferenceLinkScanner())

        # Compile the regex once
        self.replacer.compile()

    # Build a fast lookup of all files (by file name)
    def on_files(self, files: List[fl.File], config):
        self.file_mapper = FileMapper(
            options=EzLinksOptions(**self.config),
            root=config["docs_dir"],
            files=files,
            logger=LOGGER,
        )

        # After the file map has been built, initialize what we can that will
        # remain static
        self.init(config)

    def on_page_markdown(self, markdown, page, config, **kwargs):
        return self.replacer.replace(page.file.src_path, markdown)



import re
from typing import Pattern, Match

from .base_link_scanner import BaseLinkScanner
from ..types import Link, BrokenLink


class WikiLinkScanner(BaseLinkScanner):
    def pattern(self) -> str:
        # +--------------------------------+
        # | Wiki Link Regex Capture Groups |
        # +------------------------------------------------------------------------------------+
        # | wiki_is_image |  Contains ! when an image tag, or empty if not                     |
        # | wiki_link     |  Contains the Link Text between [[ wiki_link ]]                    |
        # | wiki_anchor   |  Contains the anchor, if present (e.g. file.md#anchor -> 'anchor') |
        # | wiki_text     |  Contains the text of the link.                                    |
        # +------------------------------------------------------------------------------------+
        return r"""
            (?P<wiki_is_image>[\!]?)
            \[\[
                (?P<wiki_link>[^#\|\]]*?)
                (?:\#(?P<wiki_anchor>[^\|\]]+)?)?
                (?:\|(?P<wiki_text>[^\]]+)?)?
            \]\]
            """

    def match(self, match: Match) -> bool:
        groups = match.groupdict()
        return groups.get("wiki_link") or groups.get("wiki_anchor")

    def extract(self, match: Match) -> Link:
        groups = match.groupdict()

        image = groups.get("wiki_is_image") or ""
        link = groups.get("wiki_link") or ""
        anchor = groups.get("wiki_anchor") or ""
        text = groups.get("wiki_text") or link or anchor

        if not (link or text or anchor):
            raise BrokenLink(
                f"Could not extract required field `wiki_link` from {match.group(0)}"
            )

        if anchor:
            anchor = self._slugify(anchor)

        return Link(image=image, text=text, target=link, title=text, anchor=anchor)

    def _slugify(self, link: str) -> str:
        # Convert to lowercase
        slug = link.lower()
        # Convert all spaces to '-'
        slug = re.sub(r"\ ", r"-", slug)
        # Convert all unsupported characters to ''
        return slug



import os
from typing import List
import pygtrie
import mkdocs
from pathlib import Path

from .types import EzLinksOptions
from mkdocs.structure import pages as mkpage

class FileMapper:
    def __init__(
        self,
        options: EzLinksOptions,
        root: str,
        files: List[mkpage.Page],
        logger=None,
    ):
        self.options = options
        self.root = root
        self.file_cache = {}
        self.file_trie = pygtrie.StringTrie(separator="/")
        self.logger = logger

        # Drop any files outside of the root of the docs dir
        self.files = [file for file in files if root in file.abs_src_path]

        for file in self.files:
            self._store_file(file.src_path)

    def _store_file(self, file_path):
        # Treat paths as posix format, regardless of OS
        file_path = str(Path(file_path))  # Path is a better way to normalize filepath !
        # Store the pathwise reversed representation of the file with and
        # without file extension.
        search_exprs = [file_path, os.path.splitext(file_path)[0]]
        for search_expr in search_exprs:
            # Store in fast file cache
            file_name = os.path.basename(search_expr)
            if file_name not in self.file_cache:
                self.file_cache[file_name] = [file_path]
            else:
                self.file_cache[file_name].append(file_path)
            # Store in trie
            components = list(search_expr.split("/"))
            components.reverse()
            self.file_trie["/".join(components)] = file_path

        # Reduce the dictionary to only search terms that are unique
        self.file_cache = {k: v for (k, v) in self.file_cache.items() if len(v) == 1}

    def search(self, from_file: str, file_path: str):
        abs_to = file_path
        # Detect if it's an absolute link, then just return it directly
        if abs_to.startswith("/"):
            return os.path.join(self.root, abs_to[1:])
        else:
            # Check if it is a direct link first
            from_dir = os.path.dirname(from_file)
            if os.path.exists(os.path.join(self.root, from_dir, file_path)):
                return os.path.join(self.root, from_dir, file_path)

            # It's an EzLink that must be searched
            file_name = os.path.basename(file_path)

            # Check fast file cache first
            if os.path.basename(file_name) in self.file_cache:
                abs_to = self.file_cache[file_name][0]
            else:
                search_for = list(file_path.split("/"))
                search_for.reverse()
                search_for = "/".join(search_for)

                # If we have an _exact_ match in the trie, we don't need to search
                if search_for in self.file_trie:
                    abs_to = self.file_trie[search_for]
                elif self.file_trie.has_subtrie(search_for):
                    # If we don't have an exact match, but have a partial prefix
                    values = self.file_trie.values(search_for)
                    abs_to = values[0]
                    has_ambiguity = len(values) > 1
                    # If we have ambiguities, attempt to auto-disambiguate by performing
                    # an iterative ascent of the link file's path. In this way, we should
                    # be able to get the result closest to the file doing the linking
                    if has_ambiguity:
                        file_path = os.path.dirname(from_file)
                        components = file_path.split("/")
                        components.reverse()
                        for path_component in components:
                            search_for += f"/{path_component}"
                            if (
                                self.file_trie.has_subtrie(search_for)
                                or search_for in self.file_trie
                            ):
                                new_vals = self.file_trie.values(search_for)
                                if len(new_vals) == 1:
                                    abs_to = new_vals[0]
                                    # We've resolved the ambiguity, so no need to warn
                                    has_ambiguity = False
                                    break

                    if has_ambiguity:
                        ambiguities = ""
                        for idx, file in enumerate(values):
                            active = "<--- (Selected)" if idx == 0 else ""
                            ambiguities += f"  {idx}: {file} {active}\n"
                        log_fn = (
                            self.logger.warning
                            if self.options.warn_ambiguities
                            else self.logger.debug
                        )
                        log_fn(
                            "[EzLink] Link ambiguity detected.\n"
                            f"File: '{from_file}'\n"
                            f"Link: '{search_for}'\n"
                            "Ambiguities:\n"
                            + ambiguities
                        )
        return os.path.join(self.root, abs_to)



import re
from typing import Match

from .base_link_scanner import BaseLinkScanner
from ..types import Link, BrokenLink


class MdLinkScanner(BaseLinkScanner):
    def pattern(self) -> str:
        # +------------------------------+
        # | MD Link Regex Capture Groups |
        # +-------------------------------------------------------------------------------------+
        # | md_is_image      |  Contains ! when an image tag, or empty if not (check both)      |
        # | md_alt_is_image  |  Contains ! when an image tag, or empty if not (check both)      |
        # | md_text          |  Contains the Link Text between [md_text]                        |
        # | md_protocol      |  Rejects match if link contains a protocol scheme (e.g. http)    |
        # | md_target        |  Contains the full target of the Link (filename.md#anchor)       |
        # | md_filename      |  Contains just the filename portion of the target (filename.md)  |
        # | md_anchor        |  Contains the anchor, if present (e.g. `file.md#anchor`)         |
        # | md_title         |  Contains the title, if present (e.g. `file.md "My Title"`)      |
        # +-------------------------------------------------------------------------------------+
        return r"""
            (?:
                (?P<md_is_image>\!?)\[\]
                |
                (?P<md_alt_is_image>\!?)
                \[
                    (?P<md_text>[^\]]+)
                \]
            )
            \(
                (?P<md_target>
                    (?!(?P<md_protocol>[a-z][a-z0-9+\-.]*:\/\/))
                    (?P<md_filename>\/?[^\#\ \)]*)?
                    (?:\#(?P<md_anchor>[^\)\"]*)?)?
                    (?:\ \"(?P<md_title>[^\"\)]*)\")?
                )
            \)
            """

    def match(self, match: Match) -> bool:
        return bool(match.groupdict().get("md_target"))

    def extract(self, match: Match) -> Link:
        groups = match.groupdict()
        image = groups.get("md_is_image") or groups.get("md_alt_is_image") or ""
        return Link(
            image=image,
            text=groups.get("md_text") or "",
            target=groups.get("md_filename") or "",
            title=groups.get("md_title") or "",
            anchor=groups.get("md_anchor") or "",
        )



from dataclasses import dataclass


class BrokenLink(Exception):
    print('Broken Link')
    pass


@dataclass
class Link:
    """Dataclass to hold the contents required to form a complete Link."""

    image: bool
    text: str
    target: str
    anchor: str
    title: str

    # Render as a complete MD compatible link
    def render(self):
        img = "!" if self.image else ""
        anchor = f"#{self.anchor}" if self.anchor else ""
        title = f' "{self.title}"' if self.title else ""

        return f"{img}[{self.text}]({self.target}{anchor}{title})"


@dataclass
class EzLinksOptions:
    """Dataclass to hold typed options from the configuration."""

    wikilinks: bool
    warn_ambiguities: bool
    reference_links: bool



from typing import Pattern, Match
from .base_link_scanner import BaseLinkScanner
from ..types import Link


class ReferenceLinkScanner(BaseLinkScanner):
    def pattern(self) -> str:
        # +--------------------------------------+
        # | Reference Link Regex Capture Groups  |
        # +--------------------------------------+
        # | Example: [text]: url "title"         |
        # |                                      |
        # |  text: Required                      |
        # |   url: Required                      |
        # | title: Optional, up to one newline   |
        # +--------------------------------------+
        return r"""
        (?:
        \[
            (?P<ref_text>[^\]]+)
        \]
        )\:\ 
        (?!(?P<ref_protocol>[a-z][a-z0-9+\-.]*:\/\/))
        (?P<ref_target>\/?[^\#\ \)(\r\n|\r|\n)]*)?
        (?:\#(?P<ref_anchor> [^\(\ ]*)?)?
        (?:(\r\n|\r|\n)?)?(?P<ref_title>\ ?\"[^(\r\n|\r|\n)\"]*\")?
        """

    def match(self, match: Match) -> bool:
        return bool(
            match.groupdict().get("ref_text") and match.groupdict().get("ref_target")
        )

    def extract(self, match: Match) -> Link:
        groups = match.groupdict()
        return Link(
            image=False,
            text=groups.get("ref_text") or "",
            target=groups.get("ref_target") or "",
            title=groups.get("ref_title") or "",
            anchor=groups.get("ref_anchor") or "",
        )



import os
import re
from typing import Match
from urllib.parse import quote
from .types import EzLinksOptions, BrokenLink
from .scanners.base_link_scanner import BaseLinkScanner
from .file_mapper import FileMapper


class EzLinksReplacer:
    def __init__(
        self,
        root: str,
        file_map: FileMapper,
        use_directory_urls: bool,
        options: EzLinksOptions,
        logger,
    ):
        self.root = root
        self.file_map = file_map
        self.use_directory_urls = use_directory_urls
        self.options = options
        self.scanners = []
        self.logger = logger

    def add_scanner(self, scanner: BaseLinkScanner) -> None:
        self.scanners.append(scanner)

    def replace(self, path: str, markdown: str) -> str:
        self.path = path

        # Multi-Pattern search pattern, to capture  all link types at once
        return re.sub(self.regex, self._do_replace, markdown)

    # Compiles all scanner patterns as a multi-pattern search, with
    # built in code fence skipping (individual link scanners don't
    # have to worry about them.
    def compile(self):
        patterns = "|".join([scanner.pattern() for scanner in self.scanners])
        self.regex = re.compile(
            fr"""
            (?: # Attempt to match a code block
                [`]{{3}}
                (?:[\w\W]*?)
                [`]{{3}}$
            | # Match an inline code block
                `[\w\W]*?`
            )
            | # Attempt to match any one of the subpatterns
            (?:
                {patterns}
            )
            """,
            re.X | re.MULTILINE,
        )

    def _do_replace(self, match: Match) -> str:
        abs_from = os.path.dirname(os.path.join(self.root, self.path))
        try:
            for scanner in self.scanners:
                if scanner.match(match):
                    link = scanner.extract(match)
                    # Do some massaging of the extracted results
                    if not link:
                        raise BrokenLink(
                            f"Could not extract link from '{match.group(0)}'"
                        )

                    # Handle case of local page anchor
                    if not link.target:
                        if link.anchor:
                            link.target = os.path.join(self.root, self.path)
                        else:
                            raise BrokenLink(f"No target for link '{match.group(0)}'")
                    else:
                        # Otherwise, search for the target through the file map
                        search_result = self.file_map.search(self.path, link.target)
                        if not self.use_directory_urls:
                            search_result = (
                                search_result + ".md"
                                if "." not in search_result
                                else search_result
                            )
                        if not search_result:
                            raise BrokenLink(f"'{link.target}' not found.")
                        link.target = search_result
                    link.target = quote(os.path.relpath(link.target, abs_from))
                    return link.render()
        except BrokenLink as ex:
            # Log these out as Debug messages, as the regular mkdocs
            # strict mode will log out broken links.
            self.logger.debug(f"[EzLinks] {ex}")

        # Fall through, return the original link unaltered, and let mkdocs handle it
        return match.group(0)



from setuptools import setup, find_packages

description = "A mkdocs plugin that makes linking to other documents easy."
long_description = description

version="0.2.0"

with open("README.md", "r") as f:
    long_description = f.read()
with open("requirements.txt", "r") as f:
    required=f.read().splitlines()
setup(
    name="mkdocs-ezlinked-plugin",
    version=version,
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="mkdocs, wikilinks, ezlinks, obsidian, roam",
    url="https://github.com/Mara-Li/mkdocs-ezlinks-plugin",
    author="Mara-Li",
    author_email="Mara-Li@outlook.fr",
    license="MIT",
    python_requires=">=3.6",
    install_requires=required,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
    ],
    packages=find_packages(exclude=["test.*"]),
    entry_points={
        "mkdocs.plugins": ["ezlinks = mkdocs_ezlinks_plugin.plugin:EzLinksPlugin"]
    },
)



from abc import ABCMeta, abstractmethod
from typing import Pattern, Match
from ..types import Link


class BaseLinkScanner(metaclass=ABCMeta):
    @abstractmethod
    def pattern(self) -> str:
        pass

    @abstractmethod
    def match(self, match: Match) -> bool:
        pass

    @abstractmethod
    def extract(self, match: Match) -> Link:
        pass
